# copyright 2003-2010 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of narval.
#
# narval is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# narval is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with narval.  If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-narval automatic tests"""

from logilab.common.testlib import unittest_main

from cubicweb.devtools.testlib import AutomaticWebTest


class AutomaticWebTest(AutomaticWebTest):
    no_auto_populate = set(('Plan',))
    ignored_relations = set(('execution_of',))

    def to_test_etypes(self):
        return set(('Recipe',))

    def list_startup_views(self):
        return ('narval.plan.summarytable',)


if __name__ == '__main__':
    unittest_main()



# -*- coding: utf-8 -*-
# Copyright (c) 2000-2013 LOGILAB S.A. (Paris, FRANCE).
#
# http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import os.path

from cubicweb import ValidationError, devtools

from cubes.narval.testutils import NarvalBaseTC

import narvalbot

##override default ini file
test_dir = os.path.dirname(__file__)
narvalbot._CW_SOURCES_FILE = os.path.join(test_dir, 'data', 'narval-cw-sources.ini')

class NarvalFunctionalTC(NarvalBaseTC):
    """Test for no operation recipe"""
    def test_noop(self):
        recipe = self.req.create_entity('Recipe', name=u'functest.noop',
                                        script=u'pass')
        self.run_recipe(recipe)

    def test_hello_world(self):
        "test a simple recipe is properly executed"
        recipe = self.req.create_entity('Recipe', name=u'functest.helloworld',
                                        script=u'print("hello world")')
        cwplan = self.run_recipe(recipe)
        self.assertEqual(1, len(cwplan.execution_log))
        self.assertEqual(cwplan.execution_log[0].data.read().rsplit('<br/>', 1)[-1].strip(), 'hello world')

    def test_crlf(self):
        "test a simple recipe with CRLF endlines"
        recipe = self.req.create_entity('Recipe', name=u'functest.crlf',
                                        script=u'print("hello world")\r\nprint "OK?"\r\n')
        cwplan = self.run_recipe(recipe)
        self.assertEqual(1, len(cwplan.execution_log))
        self.assertEqual(cwplan.execution_log[0].data.read().rsplit('<br/>', 1)[-1].strip(), 'hello world\nOK?')

    def test_plan_options(self):
        '''test that options defined on the `Plan` are available in
        the executed recipe, via `plan.options`'''
        recipe = self.req.create_entity('Recipe', name=u'functest.mirror',
                                        script=u'''
assert len(plan.options) == 2
assert plan.options['attr'] == '1'
assert plan.options['autre'] == '2'
''')
        self.run_recipe(recipe, options=u'attr=1\n autre=2')
        self.run_recipe(recipe, options=u'attr=1\n autre=3', expected_status='error')


    def test_traceback_in_recipe(self):
        '''check a traceback in the executed script is properly
        handled by the bot'''
        recipe = self.req.create_entity('Recipe', name=u'functest.werror',
                                        script=u'''
def toto():
    b = c

toto()
''')
        cwplan = self.run_recipe(recipe, expected_status='error')
        self.assertEqual(1, len(cwplan.execution_log))
        self.assertIn('NameError: global name',
                cwplan.execution_log[0].data.read())


    def test_sigkill(self):
        '''check the situation where the script dies via a sigkill
        (segfault or similar) is properly handled'''
        recipe = self.req.create_entity('Recipe', name=u'functest.werror',
                                        script=u'''
import os
import signal
print "banzai!"
os.kill(os.getpid(), signal.SIGKILL)
''')
        cwplan = self.run_recipe(recipe, expected_status='killed')
        self.assertEqual(1, len(cwplan.execution_log))
        self.assertIn('banzai!', cwplan.execution_log[0].data.read())

if __name__ == '__main__':
    from logilab.common.testlib import unittest_main
    unittest_main()



from cubicweb.view import EntityView, NOINDEX, NOFOLLOW
from cubicweb.predicates import is_instance, has_related_entities
from cubicweb.web import uicfg
from cubes.narval.logformat import log_to_html

def no_robot_index(self):
    return [NOINDEX, NOFOLLOW]

class FormatedLogView(EntityView):
    __select__ = EntityView.__select__ & is_instance('File')
    __regid__ = 'narval.formated_log'

    def cell_call(self, row, col, loglevel='Info', **kwargs):
        if 'dispctrl' in self.cw_extra_kwargs:
            loglevel = self.cw_extra_kwargs['dispctrl'].get('loglevel', loglevel)
        entity = self.cw_rset.get_entity(row, col)

        if entity.data:
            log_to_html(self._cw, unicode(entity.eid),
                        entity.data.read().decode(entity.data_encoding),
                        self.w, defaultlevel=loglevel)
        else:
            self.w(self._cw._('no log to display'))

_abaa = uicfg.actionbox_appearsin_addmenu
_abaa.tag_object_of(('*', 'execution_of', 'Recipe'), False)



# Copyright (c) 2013 LOGILAB S.A. (Paris, FRANCE).
#
# http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import os.path

from cubicweb import devtools

from cubes.narval.testutils import NarvalBaseTC

import narvalbot

##override default ini file
test_dir = os.path.dirname(__file__)
narvalbot._CW_SOURCES_FILE = os.path.join(test_dir, 'data', 'narval-cw-sources.ini')

class NarvalViewsTC(NarvalBaseTC):
    """Test for no operation recipe"""
    def test_plan_json(self):
        recipe = self.req.create_entity('Recipe', name=u'test.pass', script=u'pass')
        plan = self.run_recipe(recipe, options=u'a=2')
        plan.cw_clear_all_caches()
        json = self.view('ejsonexport', plan.as_rset())
        self.assertEqual(len(json), 1)
        json = json[0]
        self.assertEqual(json['eid'], plan.eid)
        self.assertEqual(json['options'], 'a=2')
        self.assertEqual(json['state'], 'done')
        self.assertEqual(json['recipe']['name'], 'test.pass')
        self.assertEqual(json['recipe']['script'], 'pass')

    def test_plan_ordering(self):
        recipe = self.req.create_entity('Recipe', name=u'test.pass', script=u'pass')
        plan_first = self.run_recipe(recipe, options=u'a=2')
        plan_ready = self.req.create_entity('Plan', execution_of=recipe)
        plan_last = self.run_recipe(recipe)
        data = self.view('narval.recipe.tab_executions', recipe.as_rset())
        plans = [b['href'] for a, b in data.a_tags if a == 'execution of test.pass']
        self.assertEqual(len(plans), 3)
        self.assertIn(str(plan_ready.eid), plans[0])
        self.assertIn(str(plan_last.eid), plans[1])
        self.assertIn(str(plan_first.eid), plans[2])

if __name__ == '__main__':
    from logilab.common.testlib import unittest_main
    unittest_main()



# Copyright (c) 2003-2013 LOGILAB S.A. (Paris, FRANCE).
# http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
"""Narval bot polling plans to be executed"""

__docformat__ = "restructuredtext en"

import os
import sys
import signal
import stat
import logging
import errno
import time
import requests
from subprocess import Popen
from tempfile import TemporaryFile
from threading import Thread, Condition, Timer, currentThread
from urlparse import urljoin, urlparse
from datetime import timedelta

from logilab.common import tasksqueue
from logilab.common.logging_ext import set_log_methods, init_log
from logilab.common.daemon import daemonize, setugid
from logilab.common.configuration import format_option_value

from . import NarvalConfiguration, options_dict, main

class NarvalTask(tasksqueue.Task):

    def __init__(self, instance_id, eid, priority=0, options=None):
        self.instance_id = instance_id
        self.planeid = int(eid)
        self.options = options_dict(options or {})
        super(NarvalTask, self).__init__("%s:%s" % (instance_id, self.planeid),
                                         -int(priority))

    def __hash__(self):
        return hash(self.id)

    def run_command(self, config):
        cmd = [sys.argv[0], 'run-plan', self.instance_id, str(self.planeid)]
        for optname, optdef in main.RunPlanCommand.options:
            value = self.options.get(optname, config[optname])
            fvalue = format_option_value(optdef, value)
            if fvalue:
                cmd.append('--'+optname)
                cmd.append(str(fvalue))
        return cmd


class NarvalProcessManager(object):
    def __init__(self, config=None):
        """ """
        self.config = config or NarvalConfiguration()
        #args = self.config.load_command_line_configuration()
        self._quiting = None
        self._running_tasks = set()
        self._cvar = Condition() # protect _running_tasks and signal free slot
        # interval of time where plan queued while there is an identical plan
        # running will be ignored
        self._skip_duplicate_time_delta = timedelta(seconds=15) # XXX

    def start(self, instance_ids=None, debug=False, pidfile=None):
        config = self.config
        self.instance_ids = instance_ids or config.cnx_infos.keys()
        if debug:
            logthreshold = config['log-threshold'] = 'DEBUG'
        else:
            logthreshold = config['log-threshold']
        init_log(debug, logthreshold=logthreshold, logfile=config['log-file'])
        self._install_sig_handlers()
        # go ! (don't daemonize in debug mode)
        if not debug and daemonize(pidfile):
            return
        # change process uid
        if config['uid']:
            setugid(config['uid'])
        self._loop()

    def _loop(self):
        """enter the service loop"""
        # service loop
        if not self.instance_ids:
            self.warning('There are no CubicWeb sources. Ensure your narval sources configuration file is correct')
            return
        while self._quiting is None:
            queue = set()
            # init the process queue
            for instance_id in self.instance_ids:
                try:
                    self.info('get pending plan from %s', instance_id)
                    for plandata in self.config.cnxh(instance_id).pending_plans():
                        self._queue_plan(queue, instance_id, plandata)
                except Exception, ex:
                    self.exception(str(ex))
            try:
                queue = sorted(queue)
                self._start_processes(queue)
            except SystemExit:
                raise
            except Exception, ex:
                self.exception('error while starting process: %s', ex)
            if not queue:
                # nothing to do, sleep a bit
                delay =  min(float(self.config.cnx_infos[instid].get('poll-delay', 0))
                             for instid in self.instance_ids) or 60
                time.sleep(delay)
            self._wait_for_free_slot()

    def _quit(self):
        """stop the server"""
        self._quiting = True
        with self._cvar:
            for _, t in self._running_tasks:
                try:
                    os.killpg(t.pid, signal.SIGTERM)
                except OSError:
                    pass
            self._cvar.notify()
        for i in xrange(5):
            time.sleep(1)
            with self._cvar:
                if not self._running_tasks:
                    return
        with self._cvar:
            for _, t in self._running_tasks:
                try:
                    os.killpg(t.pid, signal.SIGKILL)
                except OSError:
                    pass
            self._cvar.notify()

    def _install_sig_handlers(self):
        """install signal handlers"""
        self.info('installing signal handlers')
        signal.signal(signal.SIGINT, lambda x, y, s=self: s._quit())
        signal.signal(signal.SIGTERM, lambda x, y, s=self: s._quit())

    # internal process managements #############################################

    def _queue_plan(self, queue, instance_id, plan):
        self.info('queuing plan %s:%s', instance_id, plan['eid'])
        task = NarvalTask(instance_id, plan['eid'], plan['priority'],
                          plan['options'])
        with self._cvar:
            if any(task == t for _, t in self._running_tasks):
                return
        queue.add(task)

    def _wait_for_free_slot(self):
        with self._cvar:
            while not self._can_run_task() and not self._quiting:
                self._cvar.wait()

    def _can_run_task(self):
        with self._cvar:
            return len(self._running_tasks) < self.config['threads']

    def _start_processes(self, queue):
        """start pending plans according to available resources"""
        while self._can_run_task():
            try:
                self._start_process(queue.pop(0))
            except IndexError:
                break

    def _start_process(self, task):
        """start given task in a separated thread"""
        self.info('start test %s', task.id)
        # start a thread to wait for the end of the child process
        thread = Thread(target=self._run_task, args=(task,))
        with self._cvar:
            self._running_tasks.add( (thread, task) )
        thread.start()

    def _run_task(self, task):
        """run the task and remove it from running tasks set once finished"""
        try:
            self._spawn_task_process(task)
        finally:
            with self._cvar:
                self._running_tasks.remove( (currentThread(), task) )
                self._cvar.notify()
            self.info('task %s finished', task.id)

    def _spawn_task_process(self, task):
        """actually run the task by spawning a subprocess"""
        command = task.run_command(self.config)
        outfile = TemporaryFile(mode='w+', bufsize=0)
        errfile = TemporaryFile(mode='w+', bufsize=0)
        self.info(' '.join(command))
        try:
            cmd = Popen(command, bufsize=0, stdout=outfile, stderr=errfile)
        except OSError as e:
            self.error('Cannot create subprocess %r' % command)
            self.debug('Exception: %s' % e)
            return
        task.pid = cmd.pid
        maxtime = task.options.get('max-time', self.config['max-time'])
        if maxtime:
            maxreprieve = task.options.get('max-reprieve',
                                           self.config['max-reprieve'])
            maxtime = maxtime + (maxreprieve or 60) * 1.25
            timer = Timer(maxtime, os.killpg, [cmd.pid, signal.SIGKILL])
            timer.start()
        else:
            timer = None
        cmd.communicate()
        if timer is not None:
            timer.cancel()
        try:
            # kill possibly remaining children
            os.killpg(cmd.pid, signal.SIGKILL)
        except OSError, ex:
            if ex.errno != errno.ESRCH:
                raise
        for stream in (outfile, errfile):
            stream.seek(0)
        if cmd.returncode:
            self.error('error while running %s', command)
            self.error('`%s` returned with status : %s',
                       ' '.join(command), cmd.returncode)
        # add output to Plan entity and change it's state if necessary
        log = u''
        if os.fstat(errfile.fileno())[stat.ST_SIZE]:
            log += unicode(errfile.read(), 'utf-8', 'replace')
            self.info('***** %s error output', ' '.join(command))
            self.info(log)
        if os.fstat(outfile.fileno())[stat.ST_SIZE]:
            out = unicode(outfile.read(), 'utf-8', 'replace')
            self.info('***** %s standard output', ' '.join(command))
            self.info(out)
            log += out

        cnxh = self.config.cnxh(task.instance_id)
        url = cnxh.instance_url + str(task.planeid)
        if cmd.returncode:
            try:
                ret = cnxh.http_get(url, vid='fire_transition', trname='kill')
            except requests.HTTPError, ex:
                # CW returns a 409 (CONFLICT) on ValidationError,
                # expected if plan already is done or failed
                if ex.response.status_code != 409:
                    raise
        files = {'data': ('dummy', log)}
        #an empty binary object is added automatically as "data" by create_subentity
        data = cnxh.http_post(url, vid='create_subentity',
                              __cwetype__='File',
                              __cwrel__='reverse_execution_log',
                              data_name=u'execution_log.txt',
                              data_encoding='utf-8')
        if not data:
            self.warning('Could not create execution_log File for %s' % url)
        else:
            if not cnxh.http_post(url=cnxh.instance_url + 'narval-file-append',
                                  files=files, eid=data[0]['eid']):
                self.warning('Could not upload execution_log for %s' % url)


LOGGER = logging.getLogger('narval.bot')
set_log_methods(NarvalProcessManager, LOGGER)



# copyright 2003-2013 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""narval bot configuration"""

__docformat__ = "restructuredtext en"

import sys
import os, os.path as osp
import logging
from ConfigParser import ConfigParser
import hmac
try:
    import urllib3.contrib.pyopenssl
    urllib3.contrib.pyopenssl.inject_into_urllib3()
except ImportError:
    pass
import requests
from datetime import datetime
from hashlib import md5
from urlparse import urlparse

from logilab.common.configuration import Configuration, Method
from logilab.common.logging_ext import set_log_methods
from logilab.common.textutils import (splitstrip, apply_units, text_to_dict,
                                      TIME_UNITS, BYTE_UNITS)

NARVAL_SOFTWARE_ROOT = __path__[0]

if 'cubes.narval' in __name__:
    MODE = 'dev'
else:
    MODE = 'installed'

def _find_prefix(start_path=NARVAL_SOFTWARE_ROOT):
    """Runs along the parent directories of *start_path* (default to
    narval source directory) looking for one containing a
    'share/narval' directory.

    The first matching directory is assumed as the prefix installation
    of the narval bot

    Returns the matching prefix or None.
    """
    prefix = start_path
    old_prefix = None
    if not osp.isdir(start_path):
        prefix = osp.dirname(start_path)
    while (not osp.isdir(osp.join(prefix, 'share', 'narval'))
          or prefix.endswith('.egg')) and prefix != old_prefix:
        old_prefix = prefix
        prefix = osp.dirname(prefix)
    if osp.isdir(osp.join(prefix, 'share', 'narval')):
        return prefix
    return sys.prefix

try:
    INSTALL_PREFIX = os.environ['NARVAL_INSTALL_PREFIX']
except KeyError:
    INSTALL_PREFIX = _find_prefix()

class ConfigurationError(Exception):
    """a misconfiguration error"""

def options_dict(strvalue):
    options = text_to_dict(strvalue)
    for option in (u'max-cpu-time', u'max-reprieve', u'max-time'):
        if option in options:
            options[option] = apply_units(options[option], TIME_UNITS)
    if u'max-memory' in options:
        options[u'max-memory'] = apply_units(options[u'max-memory'],
                                             BYTE_UNITS)
    return options

class NarvalConfiguration(Configuration):
    options = (
        # main options
        ('uid',
         {'type' : 'string',
          'default': None,
          'help': 'if this option is set, use the specified user to start \
the daemon rather than the user running the command',
          'group': 'main', 'level': 1,
          }),

        # process control
        ('threads',
         {'type' : 'int', 'short': 'T',
          'default': 2,
          'help': 'number of plans which may be run in parallel',
          'group': 'process-control', 'level': 2,
          }),
        ('max-cpu-time',
         {'type' : 'time',
          'default': None,
          'help': 'maximum CPU Time in second that may be used to execute a test.',
          'group': 'process-control', 'level': 2,
          }),
        ('max-time',
         {'type' : 'time',
          'default': None,
          'help': 'maximum Real Time in second that may be used to execute a test.',
          'group': 'process-control', 'level': 2,
          }),
        ('max-memory',
         {'type' : 'bytes',
          'default': None,
          'help': 'maximum Memory in bytes the test can allocate.',
          'group': 'process-control', 'level': 2,
          }),
        ('max-reprieve',
         {'type' : 'time',
          'default': 60,
          'help': 'delay in second while the test try to abort nicely (reporting '
          'the error and cleaning up the environement before it\'s killed).',
          'group': 'process-control', 'level': 2,
          }),

        # server control
        ('log-file',
         {'type' : 'string',
          'default': Method('_log_file'),
          'help': 'file where output logs should be written',
          'group': 'daemon', 'level': 2,
          }),
        ('log-threshold',
         {'type' : 'string', # XXX use a dedicated type?
          'default': 'INFO',
          'help': 'narval bot\'s log level',
          'group': 'daemon', 'level': 1,
          }),
        ('poll-delay',
         {'type' : 'time',
          'default': 60,
          'help': 'delay in seconds while the server waits before looking for '
          'new pending plans ',
          'group': 'daemon', 'level': 2,
          }),
        )

    def __init__(self):
        Configuration.__init__(self)
        self.load_file_configuration(self.configuration_file)
        self.cnx_infos = connection_infos()

    if MODE == 'dev':
        _default_conf_file = osp.expanduser('~/etc/narval/narval.ini')
        _default_pid_file = '/tmp/narval.pid'
        _default_log_file = '/tmp/narval.log'
    else:
        _default_conf_file = '/etc/narval/narval.ini'
        _default_pid_file = '/var/run/narval/narval.pid'
        _default_log_file = '/var/log/narval/narval.log'

    @property
    def configuration_file(self):
        return os.environ.get('NARVALRC', self._default_conf_file)

    def _pid_file(self):
        return self._default_pid_file

    def _log_file(self):
        return self._default_log_file

    def cnxh(self, instance_id):
        return HTTPConnectionHandler(instance_id)


if MODE == 'dev':
    _DEFAULT_SOURCES_FILE = osp.expanduser('~/etc/narval/narval-cw-sources.ini')
else:
    _DEFAULT_SOURCES_FILE = '/etc/narval/narval-cw-sources.ini'

_CW_SOURCES_FILE = os.environ.get('NARVALSOURCES', _DEFAULT_SOURCES_FILE)

def connection_infos():
    if osp.exists(_CW_SOURCES_FILE):
        config = ConfigParser()
        config.read(_CW_SOURCES_FILE)
        data_sources = {}
        for section in config.sections():
            data_sources[section] = dict(config.items(section))
        return data_sources
    return {}



class HTTPConnectionHandler(object):
    """handle connection to a cubicweb repository"""

    def __init__(self, instance_id):
        if not instance_id:
            raise ConfigurationError('you must specify main cubicweb instance '
                                     'identifier in the configuration file or using '
                                     'the --cw-inst-id option')
        self.instance_id = instance_id
        self.cnxinfo = connection_infos().get(instance_id, {})
        self.instance_url = self.cnxinfo['url']
        if not self.instance_url.endswith('/'):
            self.instance_url += '/'
        self.token_id = self.cnxinfo.get('token_id')
        self.token = self.cnxinfo.get('secret')
        self.check_ssl = not self.cnxinfo.get('verify_ssl', '').lower() in ('false', 'no', '0')

    def http_get(self, url, **params):
        """GET the url and return interpreted json data"""
        return self.http_post(url, method='get', **params)

    def http_post(self, url, method='post', files=None, data=None, **params):
        """POST the url with given form parameters and files
        Keyword arguments:
        url -- url to connect to
        data -- raw data or dict containing form parameters {'key': 'value'}
        files -- files to upload {'formparam': ('filename', file_object )}
        params -- url encoded key/values pairs

        data can be raw if no file is provided otherwise it must be a
        dictionary, which will be appended as multipart content to the body.
        """
        self.debug('send HTTP %s %s params=%s, data=%s'%(method.upper(), url, params, data))
        _cw_fields = params.keys()
        if files:
            _cw_fields += files.keys()
        headers = {
                 'Accept': 'application/json',
                 'Date': datetime.utcnow().strftime('%a, %d %b %Y %H:%M:%S GMT')}
        if 'vid' in _cw_fields:
            _cw_fields.remove('vid')
        if _cw_fields:
            params['_cw_fields'] = ','.join(_cw_fields)
        try:
            r = requests.request(method, url, headers=headers,
                                 params=params, data=data, files=files, auth=self.auth,
                                 verify=self.check_ssl)
            r.raise_for_status()
        except requests.ConnectionError as e:
            self.warning('Cannot perform HTTP request: %s', e)
            return None
        try:
            return r.json()
        except ValueError as e:
            self.warning('Cannot perform HTTP request %s' % (e, ))
            self.warning('Request was url=%s, method=%s, header=%s, params=%s' %
                         (url, method, r.request.headers, params,))
            self.debug('HTTP Response: %s' % (repr(r.content),))
            return None
        except TypeError:
            # Response.json is a property in old python-requests
            return r.json

    def auth(self, req):
        if req.method in ('PUT', 'POST'):
            req.headers['Content-MD5'] = md5(req.body or '').hexdigest()
        req.headers['Authorization'] = 'Cubicweb %s:%s' % (self.token_id, self.sign(req))
        return req

    def sign(self, req):
        """Compute and return a signature for req using our secret token
        Sign the request data as expected by the signedrequest cube.
        """
        headers_to_sign = ('Content-MD5', 'Content-Type', 'Date')
        method = req.method
        url = req.url
        get_header = lambda field: req.headers.get(field, '')
        to_sign = method + url + ''.join(map(get_header, headers_to_sign))
        return hmac.new(self.token, to_sign).hexdigest()

    def pending_plans(self):
        return self.http_get(self.instance_url, vid='narval.pending-plans') or ()

    def plan(self, eid):
        url = self.instance_url + str(eid)
        data = self.http_get(url, vid='ejsonexport')
        if not len(data) == 1:
            raise Exception('expected one plan, got %s' % len(data))
        plan = data[0] # you get a list of plan of size 1
        return plan

LOGGER = logging.getLogger('narval.bot')
set_log_methods(HTTPConnectionHandler, LOGGER)




# copyright 2010 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.

__docformat__ = "restructuredtext en"
_ = unicode

from cubicweb.predicates import is_instance, none_rset
from cubicweb.web.views import tableview

from cubes.narval.views import no_robot_index

class PlanSummaryTable(tableview.EntityTableView):
    __regid__ = 'narval.plan.summarytable'
    __select__ = is_instance('Plan') | none_rset()

    title = _('Narval plans')
    category = 'startupview'

    layout_args = {'display_filter': 'top'}

    columns = ['execution', 'recipe', 'starttime', 'duration']
    column_renderers = {'execution': tableview.MainEntityColRenderer(vid='narval.plan.statuscell'),
                        'starttime': tableview.EntityTableColRenderer(lambda w, p: w(p._cw.format_date(p.starttime))),
                        'duration': tableview.EntityTableColRenderer(lambda w, p: w(unicode(p.duration))),
                        'recipe': tableview.RelatedEntityColRenderer(lambda p: p.recipe),
                        }
    html_headers = no_robot_index

    def call(self):
        self._cw.add_css('cubes.narval.css')
        if self.cw_rset is None:
            self.cw_rset = self._cw.execute(
                'Any P ORDERBY is_null(PST) DESC, PST DESC '
                'WHERE P is_instance_of Plan,'
                '      TR? wf_info_for P,'
                '      TR tr_count 0,'
                '      TR creation_date PST')
            if not self.cw_rset:
                self.w('<h1>%s</h1>' % _(self.title))
                self.w(_('no plans'))
                return
        super(PlanSummaryTable, self).call()




# copyright 2001-2010 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of Narval.
#
# Narval is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# Narval is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with Narval.  If not, see <http://www.gnu.org/licenses/>.
"""Command line input functions"""

__docformat__ = "restructuredtext en"

from logilab.common import clcommands as cl

from __pkginfo__ import version

NARVAL = cl.CommandLine('narval', doc='Control Narval agents.', version=version)
ENGINE = None # global making engine accessible **for testing purpose**

from . import NarvalConfiguration

class StartCommand(cl.Command):
    """start the narval process master. Takes optional cubicweb instance
    identifiers from which it should load its task queue.
    """
    name = 'start'
    arguments = '[<cw instance id>...]'
    options = (
        ('debug',
         {'action' : 'store_true', 'short': 'D',
          'help': 'start in debug mode',
          }),
        ('pid-file',
         {'type' : 'string',
          'help': 'daemon\'s pid file',
          }),

        )

    def run(self, args):
        from .server import NarvalProcessManager
        if self.config.debug:
            import logging
            NARVAL.logger.setLevel(logging.DEBUG)
        npm = NarvalProcessManager()
        npm.start(args, self.config.debug, self.config.pid_file)


class GenerateRCFileCommand(cl.Command):
    """generate a configuration file and exist"""
    name = 'rcfile'
    min_args = max_args = 0
    arguments = ''

    def run(self, args):
        from . import NarvalConfiguration
        nc = NarvalConfiguration()
        nc.generate_config()


class RunPlanCommand(cl.Command):
    """start a narval plan."""
    name = 'run-plan'
    arguments = '<instance id> <plan eid 1> ...'
    min_args = 2
    options = [(optname, optdef)
               for optname, optdef in NarvalConfiguration.options
               if optdef['group'] != 'daemon'
               or optname in ('log-threshold',)]

    def run(self, args):
        from . import engine
        # create Narval instance
        narval = engine.Narval(self.config)
        global ENGINE
        ENGINE = narval
        instance_id = args.pop(0)
        for eid in args:
            narval.run(instance_id, eid)


for cmdcls in (StartCommand,
               RunPlanCommand,
               #StopInstanceCommand,
               GenerateRCFileCommand,
               ):
    NARVAL.register(cmdcls)

run = NARVAL.run



# copyright 2010-2013 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-narval views/forms/actions/components for web ui"""

_ = unicode

from logilab.common.tasksqueue import PRIORITY, MEDIUM

from cubicweb import ValidationError
from cubicweb.predicates import is_instance, match_user_groups, score_entity
from cubicweb.uilib import domid
from cubicweb.view import EntityView
from cubicweb.web import Redirect
from cubicweb.web import component, formfields as ff, formwidgets as fw
from cubicweb.web.views import forms


class StartPlanForm(forms.EntityFieldsForm):
    __regid__ = 'narval.startplanform'
    __select__ = (match_user_groups('managers', 'staff')
                  & is_instance('Recipe')
                  & score_entity(lambda x: x.may_be_started()))

    form_renderer_id = 'htable'
    form_buttons = [fw.SubmitButton(label=_('start plan'))]

    @property
    def action(self):
        return self.edited_entity.absolute_url(vid='narval.startplan')

    priority = ff.IntField(choices=[(l, unicode(v)) for l, v in PRIORITY.iteritems()],
                           value=unicode(MEDIUM),
                           label=_('execution priority'),
                           sort=False, internationalizable=True)


class StartPlanComponent(component.EntityCtxComponent):
    __regid__ = 'narval.startplanform'
    __select__ = (component.EntityCtxComponent.__select__
                  & StartPlanForm.__select__)
    context = 'navcontentbottom'
    def render(self, w):
        w(u'<h2>%s</h2>' % self._cw._('start plan'))
        entity = self.cw_rset.get_entity(self.cw_row or 0, self.cw_col or 0)
        form = self._cw.vreg['forms'].select('narval.startplanform', self._cw,
                                             entity=entity)
        form.render(w=w)


class StartPlanView(EntityView):
    __regid__ = 'narval.startplan'
    __select__ = StartPlanForm.__select__
    plan = None

    def msg_url(self, msg):
        try:
            url = self._cw.build_url(self._cw.form['__redirectpath'],
                                     __message=msg)
        except KeyError:
            if len(self.cw_rset) == 1:
                if self.plan is None:
                    entity = self.cw_rset.get_entity(0, 0)
                    url = entity.absolute_url(tab=domid('narval.recipe.tab_executions'),
                                              __message=msg)
                else:
                    url = self.plan.absolute_url(__message=msg)
            else:
                url = self._cw.build_url('view', __message=msg, vid='botstatus')
        return url

    def call(self, priority=MEDIUM):
        priority = priority or self._cw.form.get('priority', MEDIUM)
        for i in xrange(self.cw_rset.rowcount):
            try:
                self.cell_call(i, 0, priority=priority)
            except ValidationError:
                raise
            except Exception, ex:
                raise Redirect(self.msg_url(unicode(ex)))
        raise Redirect(self.msg_url(self._cw._('plan(s) queued')))

    def cell_call(self, row, col, priority=MEDIUM):
        recipe = self.cw_rset.get_entity(row, col)
        self.plan = self._cw.create_entity('Plan', priority=priority,
                                           execution_of=recipe)




options = (
    # XXX only for all-in-one or repository config
    ('plan-cleanup-delay',
     {'type' : 'time',
      'default' : '60d',
      'help': ('Interval of time after which plans can be '
               'deleted. Default to 60 days. Set it to 0 if you don\'t '
               'want automatic deletion.'),
      'group': 'narval', 'level': 1,
      }),
    )

from rql.utils import register_function, FunctionDescr

class is_null(FunctionDescr):
    supported_backends = ('postgres', 'sqlite',)
    rtype = 'Boolean'

try:
    register_function(is_null)
except AssertionError:
    pass

try:
    from cubicweb.server import SQL_CONNECT_HOOKS
except ImportError: # no server installation
    pass
else:
    def init_sqlite_connection(cnx):
        def _is_null(value):
            return value is None
        cnx.create_function('IS_NULL', 1, _is_null)

    sqlite_hooks = SQL_CONNECT_HOOKS.setdefault('sqlite', [])
    sqlite_hooks.append(init_sqlite_connection)




# copyright 2010-2013 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.

__docformat__ = "restructuredtext en"
_ = unicode

from pygments import highlight
from pygments.lexers import PythonLexer
from pygments.formatters import HtmlFormatter

from logilab.common.registry import objectify_predicate

from cubicweb.predicates import is_instance, has_related_entities
from cubicweb.view import EntityView
from cubicweb.web import uicfg
from cubicweb.web.views import tabs, tableview

from cubes.narval.views import no_robot_index

_pvs = uicfg.primaryview_section
_pvdc = uicfg.primaryview_display_ctrl

# Recipe #######################################################################

_pvs.tag_attribute(('Recipe', 'name'), 'hidden')
_pvs.tag_attribute(('Recipe', 'script'), 'hidden')
_pvs.tag_object_of(('*', 'execution_of', 'Recipe'), 'hidden')

class RecipePrimaryView(tabs.TabbedPrimaryView):
    __select__ = is_instance('Recipe')

    default_tab = _('narval.recipe.tab_config')
    tabs = [default_tab, _('narval.recipe.tab_executions'),]
    html_headers = no_robot_index


class RecipeConfigTab(tabs.PrimaryTab):
    __select__ = is_instance('Recipe')
    __regid__ = 'narval.recipe.tab_config'

    def render_entity_attributes(self, entity):
        self.w(highlight(entity.script, PythonLexer(), HtmlFormatter()))
        self._cw.add_css('pygments.css')
        super(RecipeConfigTab, self).render_entity_attributes(entity)

class RecipeExecutionsTab(EntityView):
    __select__ = (is_instance('Recipe') &
                  has_related_entities('execution_of', 'object'))
    __regid__ = 'narval.recipe.tab_executions'

    html_headers = no_robot_index

    def cell_call(self, row, col):
        rset = self._cw.execute(
            'Any P ORDERBY is_null(SD) DESC, SD DESC WHERE '
            'P execution_of R, '
            'TR? wf_info_for P, '
            'TR creation_date SD, '
            'TR tr_count 0, '
            'R eid %(r)s', {'r': self.cw_rset[row][col]})
        self.wview('narval.recipe.plans_summary', rset)


@objectify_predicate
def has_errors(cls, req, rset=None, row=0, col=0, **kwargs):
    if not rset:
        return
    recipe = rset.get_entity(0, 0)
    if recipe.check_validity():
        return 1
    return 0

class RecipePlansSummaryTable(tableview.EntityTableView):
    __select__ = is_instance('Plan')
    __regid__ = 'narval.recipe.plans_summary'

    columns = ['execution', 'starttime', 'duration']
    column_renderers = {'execution': tableview.MainEntityColRenderer(),
                        'starttime': tableview.EntityTableColRenderer(lambda w, p: w(p._cw.format_date(p.starttime, time=True))),
                        'duration': tableview.EntityTableColRenderer(lambda w, p: w(unicode(p.duration)))}
    html_headers = no_robot_index

    def call(self):
        self._cw.add_css('cubes.narval.css')
        super(RecipePlansSummaryTable, self).call()




create_entity('CWGroup', name=_('narval'))
create_entity('CWUser', login=_('narval'), upassword='narval0')
rql('SET U in_group G WHERE U login "narval", G name "narval"')
rql('SET U in_group G WHERE U login "narval", G name "guests"')



# copyright 2010-2013 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-narval schema"""

_ = unicode

from logilab.common.tasksqueue import MEDIUM, PRIORITY
from yams.buildobjs import (DEFAULT_ATTRPERMS, EntityType, RelationDefinition,
                            String, Datetime, Int)
from cubicweb.schema import PUB_SYSTEM_ENTITY_PERMS, make_workflowable
from cubes.file.schema import File

IMMUTABLE_ATTR_PERMS = DEFAULT_ATTRPERMS.copy()
IMMUTABLE_ATTR_PERMS['update'] = ()

class Recipe(EntityType):
    __permissions__ = PUB_SYSTEM_ENTITY_PERMS
    name = String(maxsize=256, unique=True, required=True,
                  description=_("the recipe's name"))
    script = String(description=_("the recipe's script to execute"),
                    required=True)

class granted_permission(RelationDefinition):
    subject = 'Recipe'
    object = 'CWPermission'

class require_permission(RelationDefinition):
    subject = 'Recipe'
    object = 'CWPermission'


class Plan(EntityType):
     # add permissions actually checked according to recipe
    __permissions__ = {'read': ('managers', 'users', 'guests', 'narval'),
        'add': ('managers', 'users', 'narval'),
        'update': ('narval',),
        'delete': ('managers',),
        }

    priority = Int(default=MEDIUM,
                   vocabulary=PRIORITY.values(),
                   __permissions__=IMMUTABLE_ATTR_PERMS)
    options = String(description=_('plan option: key=value, one per line'),
                     __permissions__=IMMUTABLE_ATTR_PERMS)
    script = String(description=_('executed script'), required=True)

make_workflowable(Plan, in_state_descr=_('execution status'))

class execution_log(RelationDefinition):
    __permissions__ = {
        'read':   ('managers', 'users', 'guests', 'narval'),
        'add':    ('managers', 'narval',),
        'delete': ('managers',),
        }
    subject = 'Plan'
    object = 'File'
    cardinality = '??'
    composite = 'subject'
    inlined = True


class execution_of(RelationDefinition):
    # add permissions actually checked according to recipe
    __permissions__ = {'read': ('managers', 'users', 'guests', 'narval'),
                       'add': ('managers', 'users', 'narval'),
                       'delete': ('managers',),
                       }
    subject = 'Plan'
    object = 'Recipe'
    inlined = True
    cardinality = '1*'
    composite = 'object'

# add permissions to File objects
for perm_type in ['read', 'add', 'update']:
    File.__permissions__[perm_type] += ('narval',)



# Copyright (c) 2000-2013 LOGILAB S.A. (Paris, FRANCE).
#
# http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import os
import sys
import subprocess
import urllib2
from cStringIO import StringIO
from json import load

from logilab.common.tasksqueue import LOW

from cubicweb.devtools.testlib import CubicWebTC

try:
    import narvalbot
except:
    from cubes.narval import narvalbot
sys.modules['narvalbot'] = narvalbot

import narvalbot as bot
from narvalbot import server, main


os.killpg = lambda x,y: None


class Popen:
    pid = -1
    returncode = 0

    def __init__(self, command, bufsize, stdout, stderr):
        self.command = command
        self.stdout = stdout
        self.stderr = stderr

    def communicate(self):
        pid = os.fork()
        if not pid:
            # forked child
            try:
                sys.stdout = self.stdout
                sys.stderr = self.stderr
                main.run(self.command[1:])
            except SystemExit, ex:
                self.returncode = ex.code
            except:
                self.returncode = -1
                import traceback
                traceback.print_exc(file=sys.stderr)
        else: # parent
            pid, status = os.wait()
            self.returncode = status

class NarvalBaseTC(CubicWebTC):
    """Narval Base class for Narval Recipe test."""
    @classmethod
    def setUpClass(cls):
        super(NarvalBaseTC, cls).setUpClass()
        # monkey patch npm
        server.Popen = Popen

    @classmethod
    def tearDownClass(cls):
        super(NarvalBaseTC, cls).tearDownClass()
        server.Popen = subprocess.Popen

    def setUp(self):
        super(NarvalBaseTC, self).setUp()
        config = bot.NarvalConfiguration()

        def mp_http_post(ch, url, **form):
            if 'files' in form:
                for fname, fval in form.pop('files').items():
                    if not isinstance(fval, tuple):
                        fval = (fname, fval)
                    if isinstance(fval[1], unicode):
                        fval = (fval[0], fval[1].encode('utf-8'))
                    if isinstance(fval[1], basestring):
                        fval = (fval[0], StringIO(fval[1]))
                    form[fname] = fval
            if '_cw_fields' not in form:
                form['_cw_fields'] = ','.join([x for x in form.keys() if not x.startswith('__')])
            data, req = self.http_publish(url, form)
            fp = StringIO(data)
            if req.status_out not in (200,303):
                raise urllib2.HTTPError(url, req.status_out,
                                        "Failed", req.headers_out, fp)
            try:
                return load(fp)
            except:
                return fp

        self.orig_http_get = bot.HTTPConnectionHandler.http_get
        self.orig_http_post = bot.HTTPConnectionHandler.http_post
        bot.HTTPConnectionHandler.http_get = mp_http_post
        bot.HTTPConnectionHandler.http_post = mp_http_post
        # /end connection handler monkey patching, instantiate the process
        # manager
        self.npm = server.NarvalProcessManager(config)
        # go
        if not self.session.user.is_in_group('narval'):
            self.execute('SET U in_group G WHERE U login "admin", G name "narval"')
            self.commit()
        self.req = self.request()

    def tearDown(self):
        self.npm._quit()
        super(NarvalBaseTC, self).tearDown()
        bot.HTTPConnectionHandler.http_get = self.orig_http_get
        bot.HTTPConnectionHandler.http_post = self.orig_http_post

    def run_recipe(self, recipe, expected_status="done", options=None):
        req = recipe._cw
        cwplan = req.create_entity('Plan', priority=LOW, options=options,
                                   execution_of=recipe)
        req.cnx.commit()
        self.assertEqual(cwplan.execution_status, 'ready')
        self.run_plan(cwplan, expected_status, options)
        return cwplan

    def run_plan(self, cwplan, expected_status="done", options=None):
        queue = [server.NarvalTask('narval', cwplan.eid, cwplan.priority,
                                   cwplan.options)]
        self.npm._start_processes(queue)
        self.assertFalse(queue)
        for thread, plan in self.npm._running_tasks.copy():
            thread.join()
        cwplan = self.execute('Any X, L WHERE X eid %(x)s, X execution_log L',
                              {'x': cwplan.eid}).get_entity(0, 0)
        self.assertEqual(cwplan.execution_status, expected_status,
                         '%r (got) != %r (expected), execution log:\n%s'
                         % (cwplan.execution_status, expected_status,
                            cwplan.execution_log[0].data.getvalue()))



# copyright 2010-2012 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.

__docformat__ = "restructuredtext en"
_ = unicode

from cubicweb.view import StartupView


class BotStatusView(StartupView):
    """This view displays pending and running plans"""
    __regid__ = 'botstatus'
    title = _('Plan status')

    def call(self):
        req = self._cw
        _ = req._
        self._cw.add_css('cubes.narval.css')
        if not 'vtitle' in req.form:
            self.w(u'<h1>%s</h1>' % _(self.title))
        self._running_plans()
        self._pending_plans()
        self._latest_done_plans()

    def _running_plans(self):
        rset = self._cw.execute('Any P, P, PO WHERE P is_instance_of Plan, '
                                'P in_state S, S name "running", P options PO')
        if rset:
            self.w(u'<h2>%s</h2>'
                   % self._cw._('Running plans (%i)') % len(rset))
            self.wview('table', rset,
                       headers=[self._cw._('Plan_plural'),
                                self._cw._('options')],
                       cellvids={0: 'outofcontext', 1: 'narval.plan.optionscell'})

    def _pending_plans(self):
        rset = self._cw.execute('Any P, PR, P, PO WHERE P is_instance_of Plan, '
                                'P execution_status "ready", P priority PR, P options PO')
        if rset:
            _ = self._cw._
            self.w(u'<h2>%s</h2>' % _('Pending plans (%i)') % len(rset))
            self.wview('table', rset,
                       headers=[_('Plan_plural'), _('priority'), _('options')],
                       cellvids={0: 'outofcontext',
                                 2: 'narval.plan.optionscell'})

    def _latest_done_plans(self):
        rset = self._cw.execute('Any P,P, PO ORDERBY PMD DESC LIMIT 3 WHERE '
                                'P is_instance_of Plan, '
                                'NOT P execution_status IN ("ready", "running"), '
                                'P options PO, P modification_date PMD')
        if rset:
            self.w(u'<h2>%s</h2>' % self._cw._('Latest plans executed'))
            self.wview('table', rset,
                       headers=[self._cw._('Plan_plural'),
                                self._cw._('options')],
                       cellvids={0: 'outofcontext', 1: 'narval.plan.optionscell'})




# Copyright (c) 2000-2013 LOGILAB S.A. (Paris, FRANCE).
#
# http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
"""cubicweb-narval entity's classes"""

__docformat__ = "restructuredtext en"

from logilab.common.decorators import monkeypatch

from cubicweb.entities import AnyEntity, fetch_config
from cubicweb.entities.adapters import IFTIndexableAdapter
from cubicweb.predicates import has_related_entities, is_instance

# work around https://www.cubicweb.org/ticket/2559931
@monkeypatch(AnyEntity)
def __json_encode__(self):
    try:
        self.complete()
    except Exception:
        pass
    return super(AnyEntity, self).__json_encode__()


class Recipe(AnyEntity):
    __regid__ = 'Recipe'
    __permissions__ = ('execute',)
    rest_attr = 'name'
    fetch_attrs, cw_fetch_order = fetch_config(['name'])

    def may_be_started(self):
        return self._permissions_are_fulfiled() and not self.check_validity()

    def _permissions_are_fulfiled(self):
        if self._cw.user.matching_groups(('managers', 'narval')):
            return True
        rset = self._cw.execute('CWPermission P WHERE R require_permission P, P name "execute", '
                                'P require_group G, U in_group G, U eid %(u)s, R eid %(r)s',
                                {'u': self._cw.user.eid, 'r': self.eid})
        return bool(rset)

    def check_validity(self):
        """XXX check python script is syntaxically correct
        """
        return ()


class Plan(AnyEntity):
    __regid__ = 'Plan'
    fetch_attrs, cw_fetch_order = fetch_config(['execution_of', 'priority',
                                                'options'])

    def dc_title(self):
        return self._cw._('execution of %s') % self.recipe.dc_title()

    @property
    def recipe(self):
        return self.execution_of[0]

    @property
    def starttime(self):
        wf = self.cw_adapt_to('IWorkflowable')
        hist = wf.workflow_history
        for trinfo in hist:
            if trinfo.transition.name == 'start':
                return trinfo.creation_date

    @property
    def endtime(self):
        wf = self.cw_adapt_to('IWorkflowable')
        hist = wf.workflow_history
        for trinfo in hist:
            if trinfo.transition.name in ('end', 'kill', 'fail'):
                return trinfo.creation_date

    @property
    def duration(self):
        if self.endtime and self.starttime:
            return self.endtime - self.starttime

    @property
    def execution_status(self):
        return self.cw_adapt_to('IWorkflowable').state

    def __json_encode__(self):
        data = super(Plan, self).__json_encode__()
        data['recipe'] = self.recipe
        data['state'] = self.execution_status
        return data

class NoIndexLogFileIndexableAdapter(IFTIndexableAdapter):
    __select__ = is_instance('File') & has_related_entities('execution_log', 'object')

    def get_words(self):
        return {}



# Copyright (c) 2000-2013 LOGILAB S.A. (Paris, FRANCE).
#
# http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
"""Engine is the Narval interpreter: basically wrapped python scripts"""

from __future__ import with_statement

__docformat__ = "restructuredtext en"

import logging
import threading

from tempfile import NamedTemporaryFile

from logilab.common.proc import ResourceController, RESOURCE_LIMIT_EXCEPTION
from logilab.common import logging_ext as lext
from logilab.mtconverter import xml_escape

from . import HTTPConnectionHandler, options_dict


class HTMLFormatter(logging.Formatter):
    """A formatter for the logging standard module, using format easily
    parseable for later in the web (html) ui.
    """

    def format(self, record):
        msg = record.getMessage()
        if record.exc_info:
            # Cache the traceback text to avoid converting it multiple times
            # (it's constant anyway)
            if not record.exc_text:
                record.exc_text = self.formatException(record.exc_info)
        if record.exc_text:
            if msg[-1:] != "\n":
                msg += "\n"
            msg =  record.exc_text
        if isinstance(msg, unicode):
            msg = msg.encode('utf-8')
        return '%s\t%s\t%s\t%s<br/>' % (record.levelno,
                                        record.filename or '',
                                        record.lineno or '',
                                        xml_escape(msg))


class Narval(object):
    def __init__(self, config):
        self.rc = config
        self.resource_ctrl = ResourceController(config.max_cpu_time,
                                                config.max_time,
                                                config.max_memory,
                                                config.max_reprieve)
        threshold = lext.get_threshold(True, config.log_threshold)
        lext.init_log(True, logthreshold=threshold, fmt=HTMLFormatter())
        logging.getLogger('narval').setLevel(threshold)

    def run(self, instance_id, eid):
        """run the plan identified by its eid and exit"""
        # create cnxh
        self.cnxh = HTTPConnectionHandler(instance_id)
        # not true during functional tests due to monkey patches
        if threading.currentThread().getName() == 'MainThread':
            self.resource_ctrl.setup_limit()
        try:
            plandata = self.cnxh.plan(eid)
            self._run_plan(plandata)
        finally:
            if threading.currentThread().getName() == 'MainThread':
                self.resource_ctrl.clean_limit()

    def resource_reached(self, exc, context=None):
        """method for logger to handle RessourceError"""
        if isinstance(exc, MemoryError):
            limit = 'memory'
        else:
            limit = exc.limit
        if context is not None:
            msg = '%s resource limit reached, While %s' % (limit, context)
        else:
            msg = '%s resource limit reached' % limit
        self.critical(msg, exc_info=True)

    def _run_plan(self, plandata):
        """run the given plan"""
        narvalplan = Plan(self.cnxh, plandata)
        narvalplan.fire_transition('start')
        try:
            narvalplan.run()
            narvalplan.fire_transition('end')
        except RESOURCE_LIMIT_EXCEPTION, ex:
            self.resource_reached(ex, 'processing plan %s' % narvalplan.url)
            narvalplan.fire_transition('kill')
            raise
        except Exception:
            self.error('Exception raised while running plan %s',
                       narvalplan.url, exc_info=True)
            narvalplan.fire_transition('fail')
            raise
        except:
            narvalplan.fire_transition('fail')
            raise

LOGGER = logging.getLogger('narval.engine')
lext.set_log_methods(Narval, LOGGER)


class Plan(object):
    """a plan element is a running recipe: RecipeElement + execution context"""

    def __init__(self, cnxh, plandata):
        self.cnxh = cnxh
        self.plandata = plandata
        self.options = options_dict(plandata['options'])
        self.name = plandata['recipe']['name']
        self.script = plandata['script']
        self.url = cnxh.instance_url + str(plandata['eid'])

    def fire_transition(self, trname):
        self.info('fire transition %s of %s', trname, self.url)
        self.cnxh.http_get(self.url, vid='fire_transition', trname=trname)

    def run(self):
        with NamedTemporaryFile(suffix='.py') as f:
            f.write(self.script)
            f.flush()
            execfile(f.name, {'plan': self})

LOGGER = logging.getLogger('narval.engine')
lext.set_log_methods(Plan, LOGGER)



# coding: utf-8
"""utilities to turn raw logs into nice html reports (see format described below)

:organization: Logilab
:copyright: 2008-2010 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
:contact: http://www.logilab.fr/ -- mailto:contact@logilab.fr
:license: General Public License version 2 - http://www.gnu.org/licenses
"""
__docformat__ = "restructuredtext en"
_ = unicode

import logging
import re
from logilab.mtconverter import xml_escape
from cubicweb.utils import make_uid


REVERSE_SEVERITIES = {
    logging.DEBUG:   _('DEBUG'),
    logging.INFO:    _('INFO'),
    logging.WARNING: _('WARNING'),
    logging.ERROR:   _('ERROR'),
    logging.FATAL:   _('FATAL')
    }


# Properly formatted log are expected to use the following form:
#
#     {severity(int)}\t{filename}\t{line number}\t{msg text with potential new line}<br/>
#
# Messages that did not follow this convention are called "baldy formatted" and
# are displayed only at a debug level. However, we split the log stream on
# '<br/>' marker contained in "properly formated" log. As we can expect a
# message which didn't start properly (with the severity\tfilename\tlineno
# header) to not finish properly, it's likely that the bad ouput contained in
# the log stream does *not* end with a `<br\>` and that the `<br\>` used for
# splitting belongs to properly formated message. We can the try to recover any
# properly formatted messages embedded withing a wrongly formatted block.
#
# GOOD_MSG_STRICT_RE match the start of a perflectly formatted message. Any part
# of the stream matching it is considered a message without further processing.

# GOOD_MSG_START_RE match the start of propertly formatted message anywhere in
# block of text. (but we assume any propertly formatted message if preceded by a
# newline). We search for a match of GOOD_MSG_START_RE in the badly formatted
# block. If any match is found we split the block in two.

GOOD_MSG_STRICT_RE = re.compile(r'^[0-9]+\t.*\t.*\t')
GOOD_MSG_START_RE  = re.compile(r'^[0-9]+\t.*\t.*\t', re.MULTILINE)

def log_to_html(req, domid, data, w, defaultlevel='Info',
                severities=REVERSE_SEVERITIES):
    """format logs data to an html table

    log are encoded in the following format for each record:

      encodedmsg = u'%s\t%s\t%s\t%s<br/>' % (severity, path, line,
                                             xml_escape(msg))
    """
    req.add_css(('cubes.narval.css', 'cubicweb.tablesorter.css'))
    _selector(req, domid, w, defaultlevel, severities)
    _table_header(req, domid, w)
    _main_table(req, domid, data, w, defaultlevel, severities)

def _selector(req, domid, w, defaultlevel, severities):
    req.add_js(('cubes.narval.js', 'jquery.tablesorter.js'))
    if defaultlevel != 'Debug':
        req.add_onload('$("select.log_filter").val("%s").change();'
                       % req.form.get('log_level', defaultlevel))
    w(u'<form>')
    w(u'<label>%s</label>' % req._(u'Message Threshold'))
    w(u'<select class="log_filter" onchange="filter_log(\'%s\', this.options[this.selectedIndex].value)">'
      % domid)
    for level in [level.capitalize() for key, level in sorted(severities.items())]:
        w('<option value="%s">%s</option>' % (level, req._(level)))
    w(u'</select>')
    w(u'</form>')

def _table_header(req, domid, w):
    w(u'<table class="listing" id="%s">' % domid)
    w(u'<tr><th>%s</th><th>%s</th><th>%s</th><th>%s</th></tr>' % (
        req._('severity'), req._('path or command'), req._('line'), req._('message')))

def _main_table(req, domid, data, w, defaultlevel, severities):
    all_msg = []
    #try to cure invalid msg:
    for msg in data.split('<br/>'):
        msg = msg.strip()
        if GOOD_MSG_STRICT_RE.search(msg) is not None:
            all_msg.append(msg)
        else:
            match = GOOD_MSG_START_RE.search(msg)
            if match is not None:
                # We found some real message inside, let's save it
                pos = match.start()
                junk = xml_escape(msg[:pos])
                all_msg.append(junk)
                all_msg.append(msg[pos:])
            else:
                # hopeless junk
                all_msg.append(xml_escape(msg))
    for msg_idx, record in enumerate(all_msg):
        record = record.strip()
        if not record:
            continue
        try:
            srecord = record.split('\t', 3)
            if len(srecord) < 4: # some fields are missing, let's fill with ''
                srec = srecord
                srecord = [''] * 4
                srecord[:len(srec)] = srec
            severity, path, line, msg = srecord
            severityname = severities[int(severity)]
        except (KeyError, ValueError):
            req.warning('badly formated log %s' % record)
            path = line = u''
            severity     = logging.DEBUG
            severityname = severities[int(severity)]
            msg = record
        log_msg_id = 'log_msg_%i' % msg_idx
        w(u'<tr class="log%s" id="%s">' % (severityname.capitalize(),
                                           log_msg_id))
        w(u'<td class="logSeverity" cubicweb:sortvalue="%s">' % severity)
        data = {
            'severity': req._(severities[int(severity)]),
            'title': _('permalink to this message'),
            'msg_id': log_msg_id,
        }
        w(u'''<a class="internallink" href="javascript:;" title="%(title)s" '''
          u'''onclick="document.location.hash='%(msg_id)s';">&#182;</a>'''
          u'''&#160;%(severity)s''' % data)
        w(u'</td>')
        w(u'<td class="logPath">%s</td>' % (path or u'&#160;'))
        w(u'<td class="logLine">%s</td>' % (line or u'&#160;'))
        w(u'<td class="logMsg">')
        SNIP_OVER = 7
        lines = msg.splitlines()
        if len(lines) <= SNIP_OVER:
            w(u'<div class="rawtext">%s</div>' % msg)
        else:
            # The make_uid argument have not specific meaning here.
            div_snip_id = make_uid(u'log_snip_')
            div_full_id = make_uid(u'log_full_')
            divs_id = (div_snip_id, div_full_id)
            snip = u'\n'.join((lines[0],
                               lines[1],
                               u'  ...',
                               u'    %i more lines [double click to expand]' % (len(lines)-4),
                               u'  ...',
                               lines[-2],
                               lines[-1]))
            divs = (
                (div_snip_id, snip, u'expand', "class='collapsed'"),
                (div_full_id, u'<pre>' + msg + u'</pre>',  u'collapse', "class='hidden'")
            )
            for div_id, content, button, h_class in divs:
                text = _(button)
                js   = u"toggleVisibility('%s'); toggleVisibility('%s');" % divs_id

                w(u'<div id="%s" %s>' % (div_id, h_class))
                w(u'<div class="raw_test" ondblclick="javascript: %s" '
                   'title="%s" style="display: block;">' % (js, text))
                w(u'%s' % content)
                w(u'</div>')
                w(u'</div>')
        w(u'</td>')
        w(u'</tr>\n')
    w(u'</table>')





# Copyright (c) 2000-2013 LOGILAB S.A. (Paris, FRANCE).
#
# http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
"""cubicweb-narval validity tests classes"""

from __future__ import with_statement

__docformat__ = "restructuredtext en"

from logilab.common.tasksqueue import LOW
from cubicweb import ValidationError
from cubicweb.devtools.testlib import CubicWebTC

class NarvalSecurityTC(CubicWebTC):

    def setUp(self):
        super(NarvalSecurityTC, self).setUp()
        req = self.request()
        self.narvaux = req.create_entity('CWGroup', name=u'narvaux')
        self.toto = self.create_user(req, u'toto', groups=('narvaux', 'users'))
        self.titi = self.create_user(req, u'titi', groups=('users',))

    def test_noop(self):
        req = self.request()
        recipe = req.create_entity('Recipe', name=u'functest.noop', script=u'pass')

        eperm = req.create_entity('CWPermission', name=u'execute', label=u'execute (narvaux)')
        req.execute('SET R granted_permission P WHERE R eid %(r)s, P eid %(p)s',
                    {'r': recipe.eid, 'p': eperm.eid})
        req.execute('SET P require_group G WHERE G eid %(g)s, P eid %(p)s',
                    {'g': self.narvaux.eid, 'p': eperm.eid})

        self.commit()
        with self.login('titi'):
            req = self.request()
            self.assertRaises(ValidationError, req.create_entity, 'Plan',
                              priority=LOW,
                              execution_of=recipe)
            req.cnx.rollback()
        with self.login('toto'):
            req = self.request()
            plan = req.create_entity('Plan', priority=LOW,
                                     execution_of=recipe)
            self.commit()

if __name__ == '__main__':
    from logilab.common.testlib import unittest_main
    unittest_main()



# copyright 2010-2012 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-narval specific hooks and operations"""

from datetime import datetime, timedelta

from cubicweb import ValidationError
from cubicweb.predicates import is_instance
from cubicweb.server import hook


__docformat__ = "restructuredtext en"


class ServerStartupHook(hook.Hook):
    """add looping task to automatically start tests
    """
    __regid__ = 'narval.startup'
    events = ('server_startup',)
    def __call__(self):
        cleanupdelay = self.repo.config['plan-cleanup-delay']
        if not cleanupdelay:
            return # no auto cleanup
        cleanupinterval = min(60*60*24, cleanupdelay)
        def cleanup_plans(repo, delay=timedelta(seconds=cleanupdelay),
                          now=datetime.now):
            session = repo.internal_session()
            mindate = now() - delay
            try:
                for etype in [repo.schema['Plan']] + repo.schema['Plan'].specialized_by():
                    session.execute('DELETE %s TE WHERE '
                                    'TE modification_date < %%(min)s' % etype,
                                    {'min': mindate})
                    session.commit()
                    session.set_pool()
            finally:
                session.close()
        self.repo.looping_task(cleanupinterval, cleanup_plans, self.repo)


class AfterPlanCreated(hook.Hook):
    """a plan has been created:
    * check it can actually be started
    * if so, ask the bot to execute it

    XXX do this on the 'execution_of' relation creation
    """
    __regid__ = 'narval.start_plan'
    __select__ = hook.Hook.__select__ & is_instance('Plan')
    events = ('after_add_entity',)

    def __call__(self):
        plan = self.entity
        if not plan.recipe.may_be_started():
            msg = self._cw._('recipe may not be started')
            raise ValidationError(plan.eid, {None: msg})

class BeforePlanCreated(hook.Hook):
    __regid__ = 'narval.plan.set_script'
    __select__ = hook.Hook.__select__ & is_instance('Plan')
    events = ('before_add_entity',)

    def __call__(self):
        plan = self.entity
        if 'execution_of' not in plan.cw_edited:
            return
        eid = plan.cw_edited['execution_of']
        recipe = self._cw.execute('Recipe R WHERE R eid %(r)s',
                                  {'r': eid}).get_entity(0,0)
        plan.cw_edited['script'] = recipe.script



from cubicweb import Binary

etypelist = ['Plan'] + [i.type for i in schema['Plan'].specialized_by()]

add_cubes(('signedrequest', 'file'))

add_attribute('Recipe', 'script', commit=False)
rql('SET X script "#to be updated" WHERE X is Recipe, X script NULL')
sync_schema_props_perms(schema['Recipe'].rdef('name'))

for etype in etypelist:
    print "add script on", etype
    add_attribute(etype, 'script', commit=False)
    # we shouldn't have any, but just in case...
    rql('SET X script "unknown" WHERE X is %s, X script NULL' % etype)
commit()

drop_entity_type('RecipeStep')
drop_entity_type('RecipeTransition')
drop_entity_type('NarvalConditionExpression')
drop_entity_type('RecipeTransitionCondition')
drop_entity_type('RecipeStepInput')
drop_entity_type('RecipeStepOutput')

for rtype in ('in_recipe', 'start_step', 'end_step', 'in_steps',
              'in_error_steps', 'out_steps', 'conditions', 'takes_input',
              'generates_output', 'matches_condition'):
    if rtype not in fsschema:
        drop_relation_type(rtype)

# create the workflow for Plan and attach it to derivated entity types
if not rql('Any X, WF WHERE X name "Plan", WF workflow_of X'):
    make_workflowable('Plan')

    wf = add_workflow(u'Plan execution status workflow', 'Plan')

    ready = wf.add_state(_('ready'), initial=True)
    running = wf.add_state(_('running'))
    done = wf.add_state(_('done'))
    error = wf.add_state(_('error'),
            description=_('The execution of the Plan has been terminated (due to unexpected error)'))
    killed = wf.add_state(_('killed'),
            description=_('The execution of the Plan has been terminated (due to resource exhaustion)'))
    wf.add_transition(_('start'), ready, running, requiredgroups=('narval', 'managers'))
    wf.add_transition(_('end'), running, done, requiredgroups=('narval', 'managers'))
    wf.add_transition(_('fail'), running, error, requiredgroups=('narval', 'managers'))
    wf.add_transition(_('kill'), running, killed, requiredgroups=('narval', 'managers'))

    # attach the workflow to all derived entity types
    rql('SET WF workflow_of TE, TE default_workflow WF WHERE P default_workflow WF, '
        'P name "Plan", TE specializes P, P is CWEType')
    commit(ask_confirm=True)


# Ensure workflow status:
for etype in etypelist:
    print "Managing %s workflow" % etype
    # remove Plan (and derived CWETypes) that are in "running" state:
    # these are zombies
    rql('DELETE %s P WHERE P execution_status "running"' % etype)

    # for each TestExecution...
    te_eids = rql('DISTINCT Any X WHERE '
                  'X is %s, '
                  'NOT X in_state S' % etype)
    print "Setting %s %s in 'ready' state" % (len(te_eids), etype)
    # set it in correct initial state
    for (eid,) in te_eids:
        session.entity_from_eid(eid).cw_adapt_to('IWorkflowable').set_initial_state('ready')
    commit()
    # fire all necessary transitions
    print "Setting TestExecution states in correct state"
    to_update=[]
    for (eid,) in te_eids:
        te = session.entity_from_eid(eid)
        if te.execution_status == 'ready':
            continue

        tri = te.cw_adapt_to('IWorkflowable').fire_transition('start')
        if te.starttime:
            to_update.append((tri.eid, te.starttime))

        if te.execution_status == 'done':
            tri = te.cw_adapt_to('IWorkflowable').fire_transition('end')
        elif te.execution_status == 'killed':
            tri = te.cw_adapt_to('IWorkflowable').fire_transition('kill')
        else:
            tri = te.cw_adapt_to('IWorkflowable').fire_transition('fail')
        if te.endtime:
            to_update.append((tri.eid, te.endtime))

    # set correct timestamps
    for eid, date in to_update:
        rql('SET TI creation_date %(date)s WHERE TI eid %(eid)s',
            {'date': date, 'eid': eid})
    commit()


install_custom_sql_scripts('narval')

log_list = []

if confirm('Upgrade all execution_logs to file objects (if you say no here, execution logs will be lost)?'):
    # first create the File entities with
    from cubicweb.predicates import is_instance
    from cubicweb.entities.adapters import IFTIndexableAdapter
    class NoIndexFileIndexableAdapter(IFTIndexableAdapter):
        __select__ = is_instance('File')
        def get_words(self):
            return {}
    obj = NoIndexFileIndexableAdapter
    session.vreg.register(obj)
    registered = getattr(obj, '__registered__', None)
    if registered:
        for registry in obj.__registries__:
            registered(session.vreg[registry])
    try:
        for i, (eid,) in enumerate(rql('Any X WHERE X is_instance_of Plan, NOT X execution_log NULL')):
            data = rql('Any F WHERE X execution_log F, X eid %(eid)s', {'eid': eid})[0][0]
            rset = rql('INSERT File F: F data_name "execution_log.txt", F data %(data)s, '
                       'F data_encoding "utf-8", F data_format "text/plain"',
                       {'data': Binary(data.encode('utf-8')), 'eid': eid})

            log_list.append((eid, rset[0][0]))
            if not i%1000:
                print i, "..."
                commit(ask_confirm=False)
                print "OK"
        commit(ask_confirm=False)
    finally:
        session.vreg.unregister(obj)

for etype in etypelist:
    drop_attribute(etype, 'execution_log')
for etype in etypelist:
    add_relation_definition(etype, 'execution_log', 'File')

sync_schema_props_perms('execution_log')

# transform attribute to file
for plan_eid, log_file_eid in log_list:
        rql('SET X execution_log Y WHERE X eid %(x)s, Y eid %(y)s',
                {'x': plan_eid, 'y': log_file_eid})

commit()

for etype in etypelist:
    drop_attribute(etype, 'execution_status')
    drop_attribute(etype, 'starttime')
    drop_attribute(etype, 'endtime')
    drop_attribute(etype, 'arguments')



#!/usr/bin/env python
# pylint: disable-msg=W0404,W0622,W0704,W0613,W0152
# copyright 2003-2010 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr

__docformat__ = "restructuredtext en"

import os
import sys
import shutil
from os.path import isdir, exists, join, walk

try:
    if os.environ.get('NO_SETUPTOOLS'):
        raise ImportError()
    from setuptools import setup
    from setuptools.command import install_lib
    USE_SETUPTOOLS = 1
except ImportError:
    from distutils.core import setup
    from distutils.command import install_lib
    USE_SETUPTOOLS = 0


sys.modules.pop('__pkginfo__', None)
# import required features
from __pkginfo__ import modname, version, license, description, \
     web, author, author_email
# import optional features
import __pkginfo__
distname = getattr(__pkginfo__, 'distname', modname)
scripts = getattr(__pkginfo__, 'scripts', [])
data_files = getattr(__pkginfo__, 'data_files', None)
include_dirs = getattr(__pkginfo__, 'include_dirs', [])
ext_modules = getattr(__pkginfo__, 'ext_modules', None)
dependency_links = getattr(__pkginfo__, 'dependency_links', [])

STD_BLACKLIST = ('CVS', '.svn', '.hg', 'debian', 'dist', 'build')

IGNORED_EXTENSIONS = ('.pyc', '.pyo', '.elc', '~')

if exists('README'):
    long_description = file('README').read()
else:
    long_description = ''
if USE_SETUPTOOLS:
   requires = {}
   for entry in ("__depends__", "__recommends__"):
      requires.update(getattr(__pkginfo__, entry, {}))
   install_requires = [("%s %s" % (d, v and v or "")).strip()
                       for d, v in requires.iteritems()]
else:
   install_requires = []


def ensure_scripts(linux_scripts):
    """Creates the proper script names required for each platform
    (taken from 4Suite)
    """
    from distutils import util
    if util.get_platform()[:3] == 'win':
        scripts_ = [script + '.bat' for script in linux_scripts]
    else:
        scripts_ = linux_scripts
    return scripts_

def get_packages(directory, prefix):
    """return a list of subpackages for the given directory"""
    result = []
    for package in os.listdir(directory):
        absfile = join(directory, package)
        if isdir(absfile):
            if exists(join(absfile, '__init__.py')) or \
                   package in ('test', 'tests'):
                if prefix:
                    result.append('%s.%s' % (prefix, package))
                else:
                    result.append(package)
                result += get_packages(absfile, result[-1])
    return result

def export(from_dir, to_dir,
           blacklist=STD_BLACKLIST,
           ignore_ext=IGNORED_EXTENSIONS,
           verbose=True):
    """make a mirror of from_dir in to_dir, omitting directories and files
    listed in the black list
    """
    def make_mirror(arg, directory, fnames):
        """walk handler"""
        for norecurs in blacklist:
            try:
                fnames.remove(norecurs)
            except ValueError:
                pass
        for filename in fnames:
            # don't include binary files
            if filename[-4:] in ignore_ext:
                continue
            if filename[-1] == '~':
                continue
            src = join(directory, filename)
            dest = to_dir + src[len(from_dir):]
            if verbose:
                print >> sys.stderr, src, '->', dest
            if os.path.isdir(src):
                if not exists(dest):
                    os.mkdir(dest)
            else:
                if exists(dest):
                    os.remove(dest)
                shutil.copy2(src, dest)
    try:
        os.mkdir(to_dir)
    except OSError, ex:
        # file exists ?
        import errno
        if ex.errno != errno.EEXIST:
            raise
    walk(from_dir, make_mirror, None)

class MyInstallLib(install_lib.install_lib):
    """extend install_lib command to handle  package __init__.py and
    include_dirs variable if necessary
    """
    def run(self):
        """overridden from install_lib class"""
        install_lib.install_lib.run(self)
        # manually install included directories if any
        if include_dirs:
            base = modname
            for directory in include_dirs:
                dest = join(self.install_dir, base, directory)
                export(directory, dest, verbose=False)

def install(**kwargs):
    """setup entry point"""
    if USE_SETUPTOOLS:
        if '--force-manifest' in sys.argv:
            sys.argv.remove('--force-manifest')
    # install-layout option was introduced in 2.5.3-1~exp1
    elif sys.version_info < (2, 5, 4) and '--install-layout=deb' in sys.argv:
        sys.argv.remove('--install-layout=deb')
    if USE_SETUPTOOLS and install_requires:
        kwargs['install_requires'] = install_requires
        kwargs['dependency_links'] = dependency_links
    kwargs['packages'] = ('narvalbot',)
    #kwargs['package_dir'] = {'narvalbot' : 'narvalbot',}
    return setup(name = distname,
                 version = version,
                 license = license,
                 description = description,
                 long_description = long_description,
                 author = author,
                 author_email = author_email,
                 url = web,
                 scripts = ensure_scripts(scripts),
                 data_files = data_files,
                 ext_modules = ext_modules,
                 cmdclass = {'install_lib': MyInstallLib},
                 **kwargs
                 )

if __name__ == '__main__' :
    install()



# copyright 2010 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-narval postcreate script, executed at instance creation time or when
the cube is added to an existing instance.

You could setup site properties or a workflow here for example.
"""
_ = unicode

# create the WF of Plan
wf = add_workflow(u'Plan execution status workflow', 'Plan')

ready = wf.add_state(_('ready'), initial=True)
running = wf.add_state(_('running'))
done = wf.add_state(_('done'))
error = wf.add_state(_('error'), description=_('The execution of the Plan has been terminated (due to unexpected error)'))
killed = wf.add_state(_('killed'), description=_('The execution of the Plan has been terminated (due to resource exhaustion)'))
wf.add_transition(_('start'), ready, running, requiredgroups=('narval','managers'))
wf.add_transition(_('end'), running, done, requiredgroups=('narval','managers'))
wf.add_transition(_('fail'), running, error, requiredgroups=('narval','managers'))
wf.add_transition(_('kill'), running, killed, requiredgroups=('narval','managers'))

# create a dumb test recipe
recipe = create_entity('Recipe', name=u'test.noop', script=u'pass')




# copyright 2010-2013 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.

__docformat__ = "restructuredtext en"
_ = unicode

from tempfile import NamedTemporaryFile
from logilab.mtconverter import xml_escape
from logilab.common.tasksqueue import REVERSE_PRIORITY
from logilab.common.textutils import splitstrip

from cubicweb import tags, view, Binary
from cubicweb.predicates import (none_rset, is_instance,
                                 match_kwargs, match_form_params)
from cubicweb.web import uicfg, RequestError
from cubicweb.web.views import tabs, navigation, ibreadcrumbs, json

from cubes.narval.views import no_robot_index

from cubicweb.web.views import forms
from cubicweb.web import controller, formwidgets, formfields, Redirect

_pvs = uicfg.primaryview_section
_pvdc = uicfg.primaryview_display_ctrl

_pvs.tag_attribute(('Plan', 'execution_log'), 'relations')
_pvdc.tag_attribute(('Plan', 'execution_log'), {'vid': 'narval.formated_log'})
_pvdc.tag_attribute(('Plan', 'priority',), {'vid': 'tasksqueue.priority'})
_pvs.tag_subject_of(('Plan', 'execution_of', '*'), 'hidden')


class PlanPrimaryView(tabs.TabbedPrimaryView):
    __select__ = is_instance('Plan')

    default_tab = _('narval.plan.tab_setup')
    tabs = [default_tab]
    html_headers = no_robot_index

    def render_entity_title(self, entity):
        #self._cw.add_css('cubes.apycot.css')
        title = self._cw._('Execution of %(recipe)s') % {
            'recipe': entity.recipe.view('outofcontext')}
        self.w('<h1>%s</h1>' % title)

class PlanConfigTab(tabs.PrimaryTab):
    __regid__ = _('narval.plan.tab_setup')
    __select__ = is_instance('Plan')

    html_headers = no_robot_index


class PlanBreadCrumbTextView(ibreadcrumbs.BreadCrumbTextView):
    __select__ = is_instance('Plan')
    def cell_call(self, row, col):
        entity = self.cw_rset.get_entity(row, col)
        starttime = self._cw.format_date(entity.starttime, time=True)
        if starttime:
            self.w(starttime)
        else:
            self.w(self._cw._(entity.execution_status))

class PlanIBreadCrumbsAdapter(ibreadcrumbs.IBreadCrumbsAdapter):
    __select__ = is_instance('Plan')

    def parent_entity(self):
        """hierarchy, used for breadcrumbs"""
        return self.entity.recipe


class PlanIPrevNextAdapter(navigation.IPrevNextAdapter):
    __select__ = is_instance('Plan')

    def previous_entity(self):
        rset = self._cw.execute(
            'Any X,R ORDERBY X DESC LIMIT 1 '
            'WHERE X is Plan, X execution_of R, R eid %(c)s, '
            'X eid < %(x)s',
            {'x': self.entity.eid, 'c': self.entity.recipe.eid})
        if rset:
            return rset.get_entity(0, 0)

    def next_entity(self):
        rset = self._cw.execute(
            'Any X,R ORDERBY X ASC LIMIT 1 '
            'WHERE X is Plan, X execution_of R, R eid %(c)s, '
            'X eid > %(x)s',
            {'x': self.entity.eid, 'c': self.entity.recipe.eid})
        if rset:
            return rset.get_entity(0, 0)


class PriorityView(view.EntityView):
    __regid__ = 'tasksqueue.priority'
    __select__ = is_instance('Plan') & match_kwargs('rtype')

    def cell_call(self, row, col, rtype, **kwargs):
        entity = self.cw_rset.get_entity(row, col)
        value = getattr(entity, rtype)
        if value:
            priority = REVERSE_PRIORITY[value]
            self.w(u'<span class="priority_%s">' % priority)
            self.w(xml_escape(self._cw._(priority)))
            self.w(u'</span>')


class PlanStatusCell(view.EntityView):
    __regid__ = 'narval.plan.statuscell'
    __select__ = is_instance('Plan')

    def cell_call(self, row, col):
        entity = self.cw_rset.get_entity(row, col)
        self.w(tags.a(self._cw._(entity.execution_status),
                      href=entity.absolute_url(),
                      klass="global status_%s" % entity.execution_status,
                      title=self._cw._('see detailed execution report')))


class PlanOptionsCell(view.EntityView):
    __regid__ = 'narval.plan.optionscell'
    __select__ = is_instance('Plan')

    def cell_call(self, row, col, **kwargs):
        entity = self.cw_rset.get_entity(row, col)
        if entity.options:
            self.w(xml_escape('; '.join(entity.options.splitlines())))


class FireTransitionView(json.JsonEntityView):
    __regid__ = 'fire_transition'
    __select__ = match_form_params('trname')

    def call(self):
        self._cw.ajax_request = True
        for entity in self.cw_rset.entities():
            entity.cw_adapt_to('IWorkflowable').fire_transition(self._cw.form['trname'])
        self.wdata([])


class SetAttributes(json.JsonEntityView):
    __regid__ = 'set_attributes'

    def call(self):
        self._cw.ajax_request = True
	for entity in self.cw_rset.entities():
            kwargs = {}
            for attribute in entity.e_schema.subject_relations():
                if attribute.final and attribute.type in self._cw.form:
                    kwargs[attribute.type] = self._cw.form[attribute.type]
            if kwargs:
                entity.cw_set(**kwargs)
        self.wdata([])


class CreateSubEntity(json.JsonEntityView):
    '''
    Create a new entity from posted form:
    - `__cwetype__`: type of created entity,
    - `__cwrel__`: relation between the target entity and the created one,
    - `_cw_fields`: initialisation values for entity content

    Produces a JSON representation of the created entities.
    '''
    __regid__ = 'create_subentity'
    __select__ = (json.JsonEntityView.__select__ &
                  match_form_params('__cwetype__', '__cwrel__'))

    def call(self):
        self._cw.ajax_request = True
        etype = self._cw.form.pop('__cwetype__')
        rel = self._cw.form.pop('__cwrel__')
        subentities = []
        try:
            fields = self._cw.form['_cw_fields']
        except KeyError:
            raise RequestError(self._cw._('no edited fields specified'))
        fields = splitstrip(fields)
        kwargs = dict((k, self._cw.form[k]) for k in fields if k in
                      self._cw.form)
        kwargs.pop('vid', None)
        if etype == 'File':
            kwargs.setdefault('data', Binary())
        for entity in self.cw_rset.entities():
            kwargs[rel] = entity
            subentities.append(self._cw.create_entity(etype, **kwargs))
        self.wdata(subentities)


class PendingPlansView(json.JsonMixIn, view.StartupView):
    __regid__ = 'narval.pending-plans'
    __select__ = none_rset()

    def call(self):
        self._cw.ajax_request = True
        self.wdata([dict(eid=x, cwuri=xu, options=xo, priority=xp)
                   for x, xu, xo, xp in self._cw.execute(
            'Any X,XU,XO,XP WHERE X is_instance_of Plan, X cwuri XU, X options XO, X priority XP, '
            'X in_state S, S name "ready"')])

class FileAppendForm(forms.FieldsForm):
    __regid__ = 'narval-file-append'

    data = formfields.FileField(help=_(u'data to append'))
    eid = formfields.IntField(help=_('eid of the file entity to append to'))
    form_buttons = [formwidgets.SubmitButton()]

    def get_action(self):
        return self._cw.build_url('narval-file-append')

class FileAppendFormView(view.StartupView):
    '''
    useful for debugging, call with ?vid=narval-file-append
    '''
    __regid__ = 'narval-file-append'

    def call(self, **kwargs):
        form = self._cw.vreg['forms'].select('narval-file-append', self._cw)
        form.render(w=self.w)

class FileAppendController(controller.Controller):
    __regid__ = 'narval-file-append'

    def publish(self, rset=None):
        '''
        append to the file corresponding to the eid
        '''
        self._cw.ajax_request = True
        form = self._cw.vreg['forms'].select('narval-file-append', self._cw)
        posted = form.process_posted()
        if rset is None:
            rset = self.process_rql()
        assert len(rset) == 1
        entity = rset.get_entity(0, 0)
        log_file = self._cw.entity_from_eid(int(posted['eid']))
        with NamedTemporaryFile() as tmpfile:
            if log_file.data:
                log_file.data.to_file(tmpfile)
            while 1:
                if not posted['data']: # empty file
                    break
                buf = posted['data'].read(4096)
                if not buf:
                    break
                tmpfile.write(buf)
            tmpfile.flush()
            log_file.cw_set(data=Binary.from_file(tmpfile.name))
        raise Redirect(entity.absolute_url(vid='ejsonexport'))




# pylint: disable-msg=W0622
"""cubicweb-narval application packaging information"""

modname = 'narval'
distname = 'cubicweb-narval'

numversion = (4, 1, 0)
version = '.'.join(str(num) for num in numversion)

license = 'LGPL'
description = 'the Narval agent'
web = 'http://www.cubicweb.org/project/cubicweb-narval'
author = 'LOGILAB S.A. (Paris, FRANCE)'
author_email = 'contact@logilab.fr'

__depends__ =  {'cubicweb': '>= 3.18.0',
                'cubicweb-localperms': None,
                'cubicweb-file': '>= 1.14',
                'cubicweb-signedrequest': None,
                'Pygments': None,
                'requests': '>= 2.0.0',
               }
__recommends__ = {}


from os import listdir as _listdir
from os.path import join, isdir, exists
from glob import glob

THIS_CUBE_DIR = join('share', 'cubicweb', 'cubes', modname)
NARVAL_DIR = join('share', modname)

def listdir(dirpath):
    return [join(dirpath, fname) for fname in _listdir(dirpath)
            if fname[0] != '.' and not fname.endswith('.pyc')
            and not fname.endswith('~')
            and not isdir(join(dirpath, fname))]

data_files = [
    # common files
    [THIS_CUBE_DIR, [fname for fname in glob('*.py') if fname != 'setup.py']],
    [NARVAL_DIR, ['narvalbot/share/README',]],
    ]
# check for possible extended cube layout
for dname in ('entities', 'views', 'sobjects', 'hooks', 'schema', 'data', 'wdoc', 'i18n', 'migration'):
    if isdir(dname):
        data_files.append([join(THIS_CUBE_DIR, dname), listdir(dname)])
# Note: here, you'll need to add subdirectories if you want
# them to be included in the debian package

scripts = (join('bin', 'narval'),)
